unit interpreter;

interface
  uses MPCodeASM, numerics, sysUtils;
  Type Tsymbol = (sUnknown,
                 sConst, sNoise, sX    , sY    ,
                 sPlus , sMinus, sMul  , sDiv , sPower, sLogarithm, sMod, sMin, Max,
                 sNorm , sRound, sTrunc, sAbs , sSqr  , sSqrt, sExp, sLn,
                 sSin  , sCos  , sTan  , sASin, sACos , sATan,
                 sOpenBracket, sCloseBracket);

  Type TInstruction = (iNOP  , iPush  , iPushNoise, iPushX , iPushY,
                       iMAdd , iMSub  , iMMul     , iMDiv  , iMPow  , iMLog, iMMod  , iMMin, iMMax,
                       iMNorm, iMRound, iMTrunc   , iMAbs  ,
                       iMSqr , iMSqrt , iMExp     , iMLn   , iMSin  , iMCos  , iMTan, iMASin, iMACos, iMATan,
                       iOBracket, iCBracket);

                       {, iSMat, iXMat , iYMat, }


  Type TFormula = record
    Symbols : Array of TSymbol;
    Consts  : Array of Single;
  end;

  Type TCode = record
    Instructions : Array of TInstruction;
    Consts       : Array of Single;
  end;

  Function ParseFormula(formula : String)   : TFormula;
  Function Compile     (Formula : TFormula) : TCode;
  Function outputCompiled(code  : TCode) : String;

  Procedure executeCode(var Noise : TSingleMatrix; code : TCode);
  Procedure CompileAndExecute(var Noise : TSingleMatrix; formula : String);

  const
    cSymbols : Array[TSymbol] of ShortString = ('' , '' , 'NOISE' , 'X', 'Y',
                                                '+', '-', '*', '/', '^', 'LOG', 'MOD', 'MIN', 'MAX',
                                                'NORMALIZE', 'ROUND', 'TRUNC', 'ABS',
                                                'SQR', 'SQRT', 'EXP', 'LN',
                                                'SIN', 'COS' , 'TAN',
                                                'ARCSIN', 'ARCCOS', 'ARCTAN',
                                                '(', ')');
    cInstructions : Array[TInstruction] of ShortString =
                      ('NOP'  , 'Push'  , 'PushNoise', 'PushX' , 'PushY',
                       'MAdd' , 'MSub'  , 'MMul'     , 'MDiv'  , 'MPow' , 'MLog', 'MMod'  , 'MMin', 'MMax',
                       'MNorm', 'MRound', 'MTrunc'   , 'MAbs'  ,
                       'MSqr' , 'MSqrt' , 'MExp'     , 'MLn'   ,
                       'MSin' , 'MCos'  , 'MTan'     , 'MASin' , 'MACos', 'MATan',
                       'OBracket', 'CBracket');


implementation
Procedure CompileAndExecute(var Noise : TSingleMatrix; formula : String);
var
  Parsed : TFormula;
  Code   : TCode;
Begin
  Parsed := ParseFormula(formula);
  Code   := Compile(Parsed);
  executeCode(Noise, Code);
end;

Function outputCompiled(code : TCode) : String;
var
  i : Integer;
  c : integer;
Begin
  Result := '';
  c := 0;
  for i := 0 to high(code.Instructions) do
  Begin
    if code.Instructions[i] <> iPush then
      Result := Result + cInstructions[code.Instructions[i]] + #13#10
    else
    Begin
      Result := Result + cInstructions[code.Instructions[i]] +format(' %3f', [code.consts[c]]) + #13#10;
      inc(c);

    end;   
  end;
end;

Procedure executeCode(var Noise : TSingleMatrix; code : TCode);
var
  stack    : Array[0..64] of TSingleMatrix;
  i        : integer;
  PStack   : integer;
  pConst   : Integer;
  Operator : TMOperator;
  MFunction: TMFunction;
Begin
  PStack := 0;
  PConst := 0;

  For i := 0 to high(code.Instructions) do
  Begin
    case code.Instructions[i] of
      iNOP :
        Begin
        end;
      iPush..iPushY :
        Begin
          case code.Instructions[i] of
            iPush :
            Begin
              Stack[PStack] := SMatrix(code.consts[PConst], length(Noise), length(Noise[0]));
              inc(Pconst);
            end;
            iPushNoise : Stack[PStack] := Noise;
            iPushX     : Stack[PStack] := XMatrix(0, length(Noise), length(Noise[0]));
            iPushY     : Stack[PStack] := YMatrix(0, length(Noise), length(Noise[0]));
          end;
          inc(PStack);
        end;

      iMAdd..iMMax :
        Begin
          case code.Instructions[i] of
            iMAdd : Operator := MAdd;
            iMSub : Operator := MSub;
            iMMul : Operator := MMul;
            iMDiv : Operator := MDiv;
            iMPow : Operator := MPow;
            iMLog : Operator := MLog;
            iMMod : Operator := MMod;
            iMMin : Operator := MMin;
            iMMax : Operator := MMax;
          end;
          Operator(Stack[PStack - 2], Stack[PStack - 1]);
          Dec(PStack);
        end;
      iMNorm..iMATan :
        Begin
          case code.Instructions[i] of
            iMNorm  : MFunction := mNorm;
            iMRound : MFunction := MRound;
            iMTrunc : MFunction := MTRunc;
            iMAbs   : MFunction := MAbs;

            iMSqr   : MFunction := MSqr;
            iMSqrt  : MFunction := MSqrt;
            iMExp   : MFunction := MExp;
            iMLn    : MFunction := MLn;
            iMSin   : MFunction := MSin;
            iMCos   : MFunction := MCos;
            iMTan   : MFunction := MTan;
            iMASin  : MFunction := MASin;
            iMACos  : MFunction := MACos;
            iMATan  : MFunction := MATan;
          end;
          mFunction(Stack[PStack - 1]);
        end;
    end;
  end;
  Noise := Stack[0];
end;


Function Compile     (Formula : TFormula) : TCode;
  Procedure ReadStack(var Instructions, Stack  : Array of TInstruction;
                      var PInstr      , PStack : Integer;
                      Input                    : TInstruction);
  var
    i : integer;
  Begin
    while (PStack > 0) and ((Input <= Stack[PStack - 1]) or
            (input = iCBracket)) do
    Begin
      Instructions[PInstr] := TInstruction(Stack[PStack - 1]);
      Inc(PInstr);
      Dec(PStack);
    end;

  end;
var
  i              : TSymbol;
  n, m           : Integer;
  PInst          : Integer;
  instStack      : Array[0..15] of Array of TInstruction;
  PStack         : Array[0..15] of Integer;
  CurStack       : Integer;
  TmpInstr       : TInstruction;
Begin
  setlength(Result.Instructions, 1024);
  setlength(instStack[0], 64);
  setlength(Result.Consts, Length(Formula.consts));
  For n := 0 to high(Formula.Consts) do
    Result.Consts[n] := Formula.consts[n];


  For n := 0 to high(PStack) do
    PStack[n] := 0;
  CurStack := 0;
  PInst := 0;
  For n := 0 to high(Formula.Symbols) do
  Begin
    case Formula.Symbols[n] of
      sUnknown..sY :
      Begin
        Result.Instructions[pInst] := TInstruction(Formula.Symbols[n]);
        inc(PInst);
      end;
      sPlus..sATan  :
      Begin
        TmpInstr := TInstruction(Formula.Symbols[n]);
        ReadStack(Result.Instructions, instStack[CurStack], PInst, PStack[CurStack], TmpInstr);
        instStack[CurStack, PStack[CurStack]] := TmpInstr;
        inc(PStack[CurStack]);
      end;
      sOpenBracket :
      Begin
        Inc(CurStack);
        setlength(instStack[CurStack], 64);
      end;
      sCloseBracket :
      Begin
        ReadStack(Result.Instructions, instStack[CurStack], PInst, PStack[CurStack], iCBracket);
        setlength(instStack[CurStack], 0);
        dec(CurStack);
      end;
    end;
  end;
  ReadStack(Result.Instructions, instStack[0], PInst, PStack[0], iCBracket);

  Setlength(Result.Instructions, PInst);
End;

Function ParseFormula(formula : String) : TFormula;
var
  Position   : integer;
  i          : TSymbol;
  n          : integer;
  PSym, PCon : Byte;
  Chr        : Char;
  Str        : String;
Begin
  Position := 1;
  PSym := 0;
  PCon := 0;
  Setlength(Result.Symbols, 256);
  Setlength(Result.Consts , 256);

  repeat
    For i := sX to sPower do
      Chr := UpCase(Formula[position]);
      case chr of
        ' ' :
          Begin
          end;
       {1-Symbol Operators and consts}
        '+', '-', '*', '/', '^', '(', ')' :
          Begin
            For i := sPlus to sCloseBracket do
              If Chr = cSymbols[i] then
              Begin
                Result.Symbols[PSym] := i;
                inc(PSym);
                break;
              end;
          end;
        {numberConsts}
        '0'..'9', ',' :
        Begin
          Result.Symbols[PSym] := sConst;
          inc(PSym);
          Str := Chr;
          n := Position;
          while isnumeric(Str) do
          Begin
            inc(n);
            Str := Str + Formula[n];
          end;
          delete(Str, Length(Str), 1);
          If Length(Str) = 0 then
            Str := '0';
          position := n - 1;
          Result.Consts[PCon] := StrToFloat(Str);
          Inc(PCon);
        end;
        {comments}
        '{' :
        Begin
          while Formula[Position] <> '}' do
            inc(Position);
        end;
        'A'..'Z':
        Begin
          Str := Chr;
          n := Position;
          while (Upcase(Formula[n]) in ['A'..'Z']) and (n <= Length(Formula)) do
          Begin
            inc(n);
            Str := Str + Upcase(Formula[n]);
          end;
          delete(Str, Length(Str), 1);
          Position := n - 1;
          If Str = 'PI' then
          Begin
            Result.Symbols[PSym] := sConst;
            inc(PSym);
            Result.Consts[PCon] := pi;
            Inc(PCon);
          End
          else if Str = 'X' then
          Begin
            Result.Symbols[PSym] := sX;
            inc(PSym);
          End
          else if Str = 'Y' then
          Begin
            Result.Symbols[PSym] := sY;
            inc(PSym);
          End
          else
          Begin
            For i := sNoise to sAtan do
              If Str = cSymbols[i] then
              Begin
                Result.Symbols[PSym] := i;
                inc(PSym);
                break;
              end;
          end;
        End;
      end;
    inc(position);
  until position > length(formula);
  Setlength(Result.Symbols, PSym );
  Setlength(Result.Consts , pCon);
end;

end.
